/******************************************************  DMA Functions  *******************************************************

	Red Rapids
	797 North Grove Rd
	Suite 101
	Richardson, TX 75081
	www.redrapids.com | support@redrapids.com

	The following functions are used to control and monitor the DMA engines.

	Rev History:  08-21-2014:	Inital release.

	Copyright 2014, Red Rapids, Inc.

******************************************************  DMA Functions  *******************************************************/

#include <math.h>
#include <stdlib.h>

#include "adapter.h"
#include "sigstream.h"
#include "dma_functions.h"
#include "channel_functions.h"
#include "common_functions.h"

// Allocate DMA buffers to the requested channel number.
// Return of zero indicates success, non-zero indicates an error condition.
unsigned long DMAAllocate(s_Adapter *p_HW, unsigned long ChannelNumber, s_DMABuffer DMABuffer)
{
	unsigned long Status = 0;
	unsigned long pagecnt;
	unsigned long long descriptor;
	unsigned long add_index;

	// Set the device and channel number.
	p_HW->DMA.ChannelIndex = ChannelNumber - 1;
	// Set the burst size.
	p_HW->DMA.BurstSizeB[ChannelNumber - 1]  = DMABuffer.BurstSizeB;
	// Set the burst count.
	p_HW->DMA.BurstCount[ChannelNumber - 1]  = DMABuffer.BurstCount;
	// Set the page count.
	p_HW->DMA.PageCount[ChannelNumber - 1]   = DMABuffer.PageCount;
	// Set the pages per interrupt.
	p_HW->DMA.PagesPerMark[ChannelNumber - 1]  = DMABuffer.PagesPerMark;
	// Alocate the requested DMA buffer.
	Status = Adapter_DMABufAllocate(p_HW);
	// If allocation was successful, load the physical address of each DMA page to the hardware.
	if (Status == 0)
	{
		// Set the DMA index register to load page offsetes for the selected channel.
		Status += Write(p_HW, BAR0, DMA_DESCRIPTORS - 0x8, 0x2 <<p_HW->DMA.ChannelIndex);

		// Initialize the address index to the start of the page memory.
		add_index = DMA_DESCRIPTORS;
		// Write the DMA physical page addresses to the adapter.
		for(pagecnt=0; pagecnt < p_HW->DMA.PageCount[p_HW->DMA.ChannelIndex]; pagecnt++)
		{
			// Perform the write operation to the page memory.
			Status += Write(p_HW, BAR0, add_index, p_HW->DMA.PhysDMAAdr[p_HW->DMA.ChannelIndex][pagecnt]);
			// Increment the page memory address.
			add_index += 0x8;
		}

		// Set the DMA index register to load descriptors for the selected channel.
		Status += Write(p_HW, BAR0, DMA_DESCRIPTORS - 0x8, 0x200<<p_HW->DMA.ChannelIndex);
		
		// Initialize the address index to the start of the descriptor memory.
		add_index = DMA_DESCRIPTORS;

		// Form the descriptors and write them to the adapter.
		for(pagecnt=0; pagecnt < p_HW->DMA.PageCount[p_HW->DMA.ChannelIndex]; pagecnt++)
		{
			// Set the Burst Size and Burst Count fields of the descriptor.
			descriptor = (unsigned long long)(p_HW->DMA.BurstSizeB[p_HW->DMA.ChannelIndex] << 16) | (p_HW->DMA.BurstCount[p_HW->DMA.ChannelIndex]);

			//Set the Last Page descriptor bit if this is the final page of this channel.
			if (pagecnt == p_HW->DMA.PageCount[p_HW->DMA.ChannelIndex]-1)
				descriptor |= 0x0000000040000000ull;

			//Set the Interrupt descriptor bit if this page matches the requested pages per interrupt setting.
			if (((pagecnt+1) % p_HW->DMA.PagesPerMark[p_HW->DMA.ChannelIndex]) == 0)
				descriptor |= 0x0000000080000000ull;
			// Perform the write operation to the descriptor memory.
			Status += Write(p_HW, BAR0, add_index, descriptor);
			// Increment the descriptor memory address.
			add_index += 0x8;
		}
	}
	return(Status);
}

// Verifies that DMA configuration settings have legal values.
// Return of zero indicates no errors, non-zero indicates an error condition.
unsigned long DMAConfigCheck(s_DMABuffer DMABuffer, size_t SettingNumber)
{
	unsigned long Status = 0;
	
	switch (SettingNumber)
	{
		// Determine whether the burst size is within range and an even multiple of 8.
		case 1: {if ((DMABuffer.BurstSizeB > 16383) || fmod(DMABuffer.BurstSizeB, 16)) Status++; break;}

		// Determine whether the burst count is within range.
		case 2: {if (DMABuffer.BurstCount > 65535) Status++; break;}

		// Determine whether the page count is within range.
		case 3: {if (DMABuffer.PageCount > 1024) Status++; break;}

		// Determine whether the PagesPerMark divides evenly into the PageCount.
		case 4: {if (fmod(DMABuffer.PageCount, DMABuffer.PagesPerMark)) Status++; break;}
	}

	return Status;
}

// Report the current DMA on Demand operating status.
// Return of zero indicates no errors, non-zero indicates an error condition.
unsigned long GetDMAStatus(s_Adapter *p_HW, unsigned long ChannelNumber, s_DMAStatus *p_DMAStatus)
{
	unsigned long long Value = 0;
	unsigned long Status = 0;
	size_t cnt = 0;
	
	// Read the DMA status register.
	switch (ChannelNumber)
	{
		case 1: {Status += Read(p_HW, BAR0, DMA_1TO2_STATUS, &Value); break;}
		case 2: {Status += Read(p_HW, BAR0, DMA_1TO2_STATUS, &Value); Value >>= 32; break;}
		case 3: {Status += Read(p_HW, BAR0, DMA_3TO4_STATUS, &Value); break;}
		case 4: {Status += Read(p_HW, BAR0, DMA_3TO4_STATUS, &Value); Value >>= 32; break;}
		case 5: {Status += Read(p_HW, BAR0, DMA_5TO6_STATUS, &Value); break;}
		case 6: {Status += Read(p_HW, BAR0, DMA_5TO6_STATUS, &Value); Value >>= 32; break;}
		case 7: {Status += Read(p_HW, BAR0, DMA_7TO8_STATUS, &Value); break;}
		case 8: {Status += Read(p_HW, BAR0, DMA_7TO8_STATUS, &Value); Value >>= 32; break;}
		default: {return 0xFFFFFFFF; break;}
	}

	// Strip the current page.
	p_DMAStatus->CurrentPage = Value & 0x00000000000003FFull;

	// Strip the current burst.
	p_DMAStatus->CurrentBurst = (Value >> 16) & 0x000000000000FFFFull;

	// Strip the individual DMA engine status bits (0 = Enable, 1 = Go, 2 = Busy).
	for (cnt = 0; cnt < 2; cnt++)
		p_DMAStatus->Engine[cnt] = (Value >> (cnt + 10)) & 0x0000000000000001ull;

	// Return the total fault count.
	return(Status);
}

// Free DMA buffers that were allocated to the requested channel number.
// Return of zero indicates success, non-zero indicates an error condition.
unsigned long DMAFree(s_Adapter *p_HW, unsigned long ChannelNumber)
{
	// Set the device and channel number.
	p_HW->DMA.ChannelIndex = ChannelNumber - 1;
	// Free the requested DMA buffer.
	return(Adapter_DMABufFree(p_HW));
}